/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Bookmark Previews extension.
 *
 * The Initial Developer of the Original Code is
 * John Marshall <JohnM555@gmail.com>.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   John Marshall <JohnM555@gmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 //dump("preview service\n");
const STATE_START = Components.interfaces.nsIWebProgressListener.STATE_START;
const STATE_STOP = Components.interfaces.nsIWebProgressListener.STATE_STOP;
var PreviewService = {
  PREVIEW_WIDTH: 300,
  timeouts: [],
  listener : null,
  bookmarks: null,
  generator : null,
  isFF3 : false,
  /* since load is called around 7 times for some sites before it actually loads,
    just cache the last isBookmark check. isBookmark might not take long, but whatever */
  _lastCheck : {uri : {spec : null}, isBookmarked : false},
  onLoad: function() {
    var appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
                        .getService(Components.interfaces.nsIXULAppInfo);
    var versionChecker = Components.classes["@mozilla.org/xpcom/version-comparator;1"]
                                   .getService(Components.interfaces.nsIVersionComparator);
    PreviewService.isFF3 = (versionChecker.compare(appInfo.version, "3.0a") >= 0);
    if (PreviewService.isFF3)
      PreviewService.bookmarks = new PreviewServiceFF3();
    else{
      /*
        Use the listener to check for redirected bookmarks.
        FF3 doesn't need this because it has a better bookmarking system.
      */
      PreviewService.bookmarks = new PreviewServiceFF2();
      PreviewService.listener = new BookmarkPreviewsProgressListener();
      PreviewService.listener.onURILoad = PreviewService.onURILoad;
      PreviewService.listener.getBrowser = PreviewService.getBrowserForURI;
      gBrowser.addProgressListener(PreviewService.listener,Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
    }
    gBrowser.addEventListener('load',PreviewService.onPageLoad,true);
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService);
    prefs = prefs.getBranch("extensions.bookmarkpreviews.");
    if (!prefs.prefHasUserValue("firstrun")){
      PreviewService.showFirstRunDialog();
      prefs.setBoolPref("firstrun",true);
    }
  },
  onUnload: function() {
    gBrowser.removeEventListener('load',PreviewService.onPageLoad,true);
    if (!PreviewService.isFF3)
      gBrowser.removeProgressListener(PreviewService.listener);
    PreviewService.bookmarks.unload();
  },
  onPageLoad : function(event){
    var doc = event.originalTarget;
    if (doc instanceof Components.interfaces.nsIDOMDocument) {
      if (doc.defaultView.frameElement) {
        // Frame within a tab was loaded. doc should be the root document of
        // the frameset.
        // Find the root document:
        while (doc.defaultView.frameElement) {
          doc=doc.defaultView.frameElement.ownerDocument;
        }
      }
      if (doc && doc.URL){
        var uri = makeURI(doc.URL);
        if (!uri) return;
        var browser = gBrowser.getBrowserForDocument(doc);
        if (browser)
          PreviewService.onURILoad(uri, browser);
      }
    }
  },
  onURILoad : function(uri,browser){
    var isBookmarked = uri.equals(PreviewService._lastCheck.uri)?PreviewService._lastCheck.isBookmarked:PreviewService.bookmarks.isBookmarked(uri);
    //dump("onuriload: "+isBookmarked+"\n");
    if (isBookmarked){
      if (PreviewService.timeouts[uri.spec]!=null)
        clearTimeout(PreviewService.timeouts[uri.spec]);
      /* Just to give it time to load a little more... */
      PreviewService.timeouts[uri.spec]=setTimeout(function(){
        PreviewService.timeouts[uri.spec]=null;
        PreviewService.createPreviewFromBrowser(browser,uri);
      },1500);
    }
    PreviewService._lastCheck.uri = uri;
    PreviewService._lastCheck.isBookmarked = isBookmarked;
  },
  getBrowserForURI : function(uri){
    var numTabs = gBrowser.tabContainer.childNodes.length;
    for(var index=0; index<numTabs; index++) {
      var currentBrowser = gBrowser.getBrowserAtIndex(index);
      if (uri.equals(currentBrowser.currentURI)) {
        return currentBrowser;
      }
    }
    return null;
  },
  bookmarkAdded : function(uri){
    var currentBrowser = this.getBrowserForURI(uri);
    if (currentBrowser)
      this.createPreviewFromBrowser(currentBrowser);
  },
  createPreviewFromBrowser : function(browser, uri, aWidth, aHeight){
    uri = uri ? uri : browser.currentURI;
    //dump("creatpreviewfrombrowser: "+uri.spec+"\n");
    var win=browser.contentWindow;
    aWidth = (aWidth && aWidth!=0)?aWidth:win.innerWidth;
    aHeight = (aHeight && aHeight!=0)?aHeight:win.innerHeight;
    var canvas = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
    var scale=this.PREVIEW_WIDTH/(aWidth);
    canvas.height = Math.floor(scale*(aHeight));
    canvas.width = this.PREVIEW_WIDTH;
    var ctx = canvas.getContext("2d");
    ctx.save();
    ctx.scale(scale, scale);
    ctx.drawWindow(win, 0, 0, aWidth, aHeight,"rgb(255,255,255)");
    ctx.restore();
    var data = canvas.toDataURL("image/jpeg", "");
    if(data && data != "") {
      try {
        this.bookmarks.savePreview(data,uri);
        this.sendUpdateNotification(uri);
      } catch(e) {
        Components.utils.reportError("Bookmarks Previews: Error saving bookmark image for uri\n"+uri.spec+"\nError: "+e+"\n");
      }
    }
  },
  sendUpdateNotification : function(uri){
    /* When notifying the observers right away, the observers still seem to think
      that the file doesn't exist. */
    setTimeout(function(){
      Components.classes["@mozilla.org/observer-service;1"]
           .getService(Components.interfaces.nsIObserverService)
           .notifyObservers(uri, "bookmarkpreviews:preview-updated", null);
    },0);
  },
  showFirstRunDialog : function(){
    var self = this;
    var strbundleService = Components.classes ["@mozilla.org/intl/stringbundle;1"]
                        .getService(Components.interfaces.nsIStringBundleService);
    this.bundle = strbundleService.createBundle("chrome://bookmarkpreviews/locale/strings.properties");
    var dialogQ = this.bundle.GetStringFromName("firstrundialog");
    var dialogTitle = this.bundle.GetStringFromName("firstrundialogtitle");
    /* Give it some time for the browser to actually load */
    setTimeout(function(){
      var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                        .getService(Components.interfaces.nsIPromptService);
      var result = prompts.confirm(window, dialogTitle, dialogQ);
      if (result){
        self.generateAllPreviews();
      } else
        delete this.bundle;
    },2000);
  },
  generateAllPreviews : function(){
    //dump("generate all previews\n");
    if (!this.generator){
      this.generator = new PreviewGenerator();
      this.generator.generateAllPreviews();
    }
  }
}
function PreviewGenerator(){
}
PreviewGenerator.prototype = {
  LOAD_TIMEOUT : 1000 * 60,
  generateAllPreviews : function(){
    //dump("generateallpreviews\n");
    this.allBookmarks = PreviewService.bookmarks.getAllBookmarks();
    //dump("allbookmarks: "+this.allBookmarks.length+"\n");
    if (!this.allBookmarks || this.allBookmarks.length==0){
      delete this.allBookmarks;
      return;
    }
    if (!this.bundle){
      var strbundleService = Components.classes ["@mozilla.org/intl/stringbundle;1"]
                        .getService(Components.interfaces.nsIStringBundleService);
      this.bundle = strbundleService.createBundle("chrome://bookmarkpreviews/locale/strings.properties");
    }
    this.curPreview = -1;
    var browser = document.createElement("browser");
    browser.setAttribute("id","bookmarkpreviews-browser");
    browser.setAttribute("width",gBrowser.boxObject.width);
    browser.setAttribute("height",1);
    browser.setAttribute("type","content");
    //browser.setAttribute("height",gBrowser.boxObject.height);
    //browser.setAttribute("collapsed",true);
    document.documentElement.appendChild(browser);

    /* The canvas can't save flash/java so don't waste time loading it*/
    browser.docShell.allowPlugins = false;
    /* So it doesn't launch any windows as its working in the background*/
    browser.docShell.allowJavascript = false;
    browser.stop();

    var listener = new BookmarkPreviewsProgressListener();
    listener.getBrowser = function BP_GetBrowser(uri){
      return document.getElementById("bookmarkpreviews-browser");
    }
    listener.onURILoad = this.generatePreviewLoad;
    browser.addProgressListener(listener,Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
    browser.addEventListener("load",this,true);
    this.previewGenListener = listener;
    this.loadNextBookmark(browser);
  },
  generatePreviewLoad : function(uri, browser){
    //dump("preview load: "+uri+"\n");
    var self = PreviewService.generator;
    if (uri.equals(self.allBookmarks[self.curPreview])){
      if (self.generateTimeout!=null){
        //dump("clear timeout: "+self.generateTimeout+"\n");
        clearTimeout(self.generateTimeout);
      }
      else{
        //dump("no existing generate timeout: "+self.generateTimeout+"\n");
        //dump("pre url: "+uri.spec+"\n");
      }
      self.generateTimeout=setTimeout(function(){
        self.generateTimeout = null;
        PreviewService.createPreviewFromBrowser(browser,uri,null,gBrowser.boxObject.height);
        browser.stop();
        self.loadNextBookmark(browser);
      },1500);
    }
  },
  handleEvent : function(event){
    if (event.type == "load")
      this.onPageLoad(event);
  },
  onPageLoad : function(event){
    var doc = event.originalTarget;
    if (doc instanceof Components.interfaces.nsIDOMDocument) {
      if (doc.defaultView.frameElement) {
        // Frame within a tab was loaded. doc should be the root document of
        // the frameset.
        // Find the root document:
        while (doc.defaultView.frameElement) {
          doc=doc.defaultView.frameElement.ownerDocument;
        }
      }
      try{
      var uri = makeURI(doc.URL);
      if (!uri) return;
      var browser = document.getElementById("bookmarkpreviews-browser");
      this.generatePreviewLoad(uri, browser);
      }catch(e){
        Components.utils.reportError(e);
      }
    }
  },
  loadNextBookmark : function(browser){
    var uri = null;
    while (!uri){
      this.curPreview++;
      if (this.curPreview>=this.allBookmarks.length){
        break;
      } else {
        uri = this.allBookmarks[this.curPreview];
        if (uri && PreviewService.bookmarks.hasPreview(uri.spec))
          uri = null;
      }
    }
    if (!uri)
      this.generatePreviewsComplete(browser);
    else{
      //var time = new Date().getTime(); dump("uri: "+uri.spec); if (this.lastTime) dump(" time: "+(time-this.lastTime)+"\n"); else dump("\n"); this.lastTime = time;
      var statusTextFld = document.getElementById("statusbar-display");
      var statustext = this.bundle.GetStringFromName("creatingpreview.statustext");
      statusTextFld.label = statustext.replace("$1",(this.curPreview+1)).replace("$2",this.allBookmarks.length).replace("$3",uri.spec)
      //browser.loadURI(uri.spec);
      var flags = Ci.nsIWebNavigation.LOAD_FLAGS_BYPASS_HISTORY;
      browser.loadURIWithFlags(uri.spec,flags,null,null);
      var self = this;
      this.generateTimeout = setTimeout(function(){
        self.generateTimeout = null;
        //dump("generating preview timed out\n");
        //browser = document.getElementById("bookmarkpreviews-browser");
        browser.stop();
        self.loadNextBookmark(browser);
      },this.LOAD_TIMEOUT);
    }
  },
  generatePreviewsComplete : function(browser){
    browser.removeProgressListener(this.previewGenListener);
    browser.removeEventListener("load",this,true);
    browser.parentNode.removeChild(browser);
    delete this.previewGenListener;
    delete this.allBookmarks;
    delete this.curPreview;
    delete this.bundle;
    alert("Bookmark Previews Generated");
  }
}
function BookmarkPreviewsProgressListener(){

}
BookmarkPreviewsProgressListener.prototype = {
  onURILoad : null,
  getBrowser : null,
  QueryInterface: function(aIID) {
   if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
       aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
       aIID.equals(Components.interfaces.nsISupports))
     return this;
   throw Components.results.NS_NOINTERFACE;
  },
  onStateChange: function(aProgress, aRequest, aFlag, aStatus) {
    if(aFlag & STATE_STOP){
      if (aRequest instanceof Components.interfaces.nsIChannel){
        var chan = aRequest.QueryInterface(Components.interfaces.nsIChannel);
        if (!chan.URI.equals(chan.originalURI)){
          //dump("redirected from "+chan.originalURI.spec+" to "+chan.URI.spec+"\n");
          var browser = this.getBrowser(chan.URI);
          if (browser)
            this.onURILoad(chan.originalURI,browser);
        }
      }
    }
    return 0;
  },
  onLocationChange: function(aProgress, aRequest, aURI){},
  onProgressChange: function() {return 0;},
  onStatusChange: function() {return 0;},
  onSecurityChange: function() {return 0;},
  onLinkIconAvailable: function() {return 0;}
}
function PreviewServiceFF3(){
  PlacesUtils.bookmarks.addObserver(this, false);
  this.dir = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  this.dir.append("bookmarkpreviews");
  if (!this.dir.exists())
    this.dir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE,0777);
}
PreviewServiceFF3.prototype = {
  toString : function(){return "Firefox 3 Preview Service";},
  isBookmarked : function(uri){
    return PlacesUtils.bookmarks.getBookmarkedURIFor(uri)!=null;
  },
  hasPreview : function(aUrl){
    return this._getPreviewFile(aUrl).exists();
  },
  _getPreviewFile : function(aURL){
    var file = this.dir.clone();
    file.append(encodeURIComponent(aURL.toLowerCase())+".jpg");
    return file;
  },
  savePreview : function(data,uri){
    var file = this._getPreviewFile(uri.spec);
    var io = Components.classes["@mozilla.org/network/io-service;1"]
                       .getService(Components.interfaces.nsIIOService);
    var source = io.newURI(data, "UTF8", null);
    var target = io.newFileURI(file)
    // prepare to save the data
    var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                            .createInstance(Components.interfaces.nsIWebBrowserPersist);
    persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;

    persist.saveURI(source, null, null, null, null, file);
  },
  unload : function(){
    PlacesUtils.bookmarks.removeObserver(this,false);
  },
  /* nsINavBookmarkObserver */
  onBeginUpdateBatch : function(){},
  onEndUpdateBatch : function(){},
  onItemAdded : function(aItemId, aFolder, aIndex){
    var uri = PlacesUtils.bookmarks.getBookmarkURI(aItemId);
    PreviewService.bookmarkAdded(uri);
  },
  onItemRemoved : function(aItemId,  aFolder,  aIndex){
    //PlacesUtils.annotations.removeItemAnnotation(aItemId,"bookmark/preview");
    var uri = PlacesUtils.bookmarks.getBookmarkURI(aItemId);
    if (!this.isBookmarked(uri)){
      var file = this._getPreviewFile(uri.spec);
      if (file.exists())
        file.remove(false);
    }
  },
  onItemChanged : function(aBookmarkId, aProperty, aIsAnnotationProperty, aValue){},
  onItemVisited : function(aBookmarkId,  aVisitID, time){},
  onItemMoved : function(aItemId,  aOldParent,  aOldIndex,  aNewParent,  aNewIndex){},
  createInstance: function (outer, iid) {
    if (outer != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return this.QueryInterface(iid);
  },

  QueryInterface: function PS_QueryInterface(aIID) {
    if (aIID.equals(Ci.nsINavBookmarkObserver) ||
         aIID.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_NOINTERFACE;
  },

  /* generate all previews... */
  getAllBookmarks : function(root,bkmks){
    if (!root){
      root = Application.bookmarks;
      bkmks = [];
    }
    root.children.forEach(function(item,idx){
      if (item.type=="bookmark"){
        bkmks.push(item.uri);
      }
      else if (item.type=="folder"){
        this.getAllBookmarks(item,bkmks);
      }
    },this);
    return bkmks;
  }
}
function PreviewServiceFF2(){
  initServices();
  initBMService();
  this.bmsvc = BMSVC;
  var bkmkTxnSvc = Components.classes["@mozilla.org/bookmarks/transactionmanager;1"]
            .getService(Components.interfaces.nsIBookmarkTransactionManager);
  bkmkTxnSvc.transactionManager.AddListener(this);

  this.dir = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  this.dir.append("bookmarkpreviews");
  if (!this.dir.exists())
    this.dir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE,0777);
}
PreviewServiceFF2.prototype = {
  toString : function(){return "Firefox 3 Preview Service";},
  unload : function(){
    var bkmkTxnSvc = Components.classes["@mozilla.org/bookmarks/transactionmanager;1"]
            .getService(Components.interfaces.nsIBookmarkTransactionManager);
    bkmkTxnSvc.transactionManager.RemoveListener(this);
  },
  isBookmarked : function(uri){
    var NC = "http://home.netscape.com/NC-rdf#";
    var RDF = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
   	var BMDS  = RDF.GetDataSource("rdf:bookmarks");
    var urlArc = RDF.GetResource(NC+"URL");
    var elements = BMDS.GetAllResources();
    while (elements.hasMoreElements()){
      var ele = elements.getNext();
      if(BMDS.hasArcOut(ele, urlArc)){
        var target = BMDS.GetTarget(ele, urlArc, true);
        var value = target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
        //dump("val: "+value+"|"+uri.spec+"\n");
        if(value == uri.spec){
            //dump("isBookmark - true - "+uri.spec+"\n");
            return true
        }
      }
    }
    //dump("isBookmark - false - "+uri.spec+"\n");
    return false;
  },
  hasPreview : function(aUrl){
    return this._getPreviewFile(aUrl).exists();
  },
  _getPreviewFile : function(aURL){
    var file = this.dir.clone();
    file.append(encodeURIComponent(aURL.toLowerCase())+".jpg");
    return file;
  },
  savePreview : function(data,uri,aArg){
    var file = this._getPreviewFile(uri.spec);
    var io = Components.classes["@mozilla.org/network/io-service;1"]
                       .getService(Components.interfaces.nsIIOService);
    var source = io.newURI(data, "UTF8", null);
    var target = io.newFileURI(file)
    // prepare to save the data
    var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                            .createInstance(Components.interfaces.nsIWebBrowserPersist);
    persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;

    persist.saveURI(source, null, null, null, null, file);
  },
  removePreview : function(url){
    if (!this.isBookmarked(makeURI(url))){
      var file = this._getPreviewFile(url);
      if (file.exists())
        file.remove(false);
    }
  },
  observeTransaction : function(aTxn, aDo){
    if (!aTxn) return;
    aTxn = aTxn.wrappedJSObject;
    var type = aTxn.type;
    var item = aTxn.item;
    var itemType = BookmarksUtils.resolveType(item);
    var itemURL = null;
    if (itemType == "Bookmark"){
      var urlProp = RDF.GetResource("http://home.netscape.com/NC-rdf#URL");
      item.QueryInterface(kRDFRSCIID);
      itemURL = BMDS.GetTarget(item, urlProp, true);
      if (itemURL){
        itemURL = itemURL.QueryInterface(kRDFLITIID).Value;
        if ((type == "insert" && aDo) || (type == "remove" && !aDo)){
          PreviewService.bookmarkAdded(makeURI(itemURL));
        }
        else if ((type == "remove" && aDo) || (type == "insert" && !aDo)){
          this.removePreview(itemURL);
        }
      }
    }
  },
  /* Transaction Listener */
  willDo: function (aTxmgr, aTxn) {},
  didDo : function (aTxmgr, aTxn) {
    this.observeTransaction(aTxn, true);
  },
  willUndo: function (aTxmgr, aTxn) {},
  didUndo : function (aTxmgr, aTxn) {
    this.observeTransaction(aTxn, false);
  },
  willRedo: function (aTxmgr, aTxn) {},
  didRedo : function (aTxmgr, aTxn) {
    this.observeTransaction(aTxn, true);
  },
  didMerge       : function (aTxmgr, aTxn) {},
  didBeginBatch  : function (aTxmgr, aTxn) {},
  didEndBatch    : function (aTxmgr, aTxn) {
    this.observeTransaction(aTxn, true);
  },
  willMerge      : function (aTxmgr, aTxn) {},
  willBeginBatch : function (aTxmgr, aTxn) {},
  willEndBatch   : function (aTxmgr, aTxn) {},

  /* Used for generating all previews */
  getAllBookmarks : function(){
    var children = [], all = [];
    BookmarksUtils.getAllChildren(RDF.GetResource("NC:BookmarksRoot"),children);
    for (var i = 0; i < children.length; i++){
      try{
        if (children[i][1] && children[i][1].Value!=null){
          all.push(makeURI(children[i][1].Value));
        }
      }catch(e){dump(e+"\n");}
    }
    return all;
  }
}
window.addEventListener('load',function(e) { PreviewService.onLoad(e); },false);
window.addEventListener('unload',function(e) { PreviewService.onUnload(e); },false);
